import numpy as np 

from .TopoMap import TopoMap

from .UnionFindComponents import UnionFind

class TopoTree(TopoMap):
    """Generate TopoTree representation of the input data points.

    Parameters
    ----------

    metric : {'cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan',' braycurtis', 'canberra', \
                'chebyshev', 'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis', \
                'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath',\
                'sqeuclidean', 'yule'} or a callable, default='euclidean'
        Distance metric to compute the MST. Check out sklearn's pairwise_distances to see other options.

    approach: {'mlpack', 'nx', 'ANN'}, default='mlpack'
        Approach for computing the MST. If 'mlpack', mlpack's function emst is used. In this case, the only 
        metric available is euclidean. If 'nx', it uses sklearn's pairwise_distance function followed by 
        networkx's minimum_spanning_tree function. If 'ANN', it uses DiskANN's precomputed approximate 
        nearest neighbor instead of computing a MST. In this case, you need to provide the index_path.

    load_mst: Bool, default=False
        Option to load precomputed mst. If set to True, you need to provide the mst_path.

    mst_path: string, default=''
        Path to the precomputed mst. Only used if load_mst is True.

    index_path: string, default=''
        Path to index file generated by DiskANN. Only used if approach is set to 'ANN'.

    min_box_size: int, default=1
        Minumum number of points for a component to have an assigned box.
    """

    def __init__(self, 
                 metric='euclidean',
                 approach = 'mlpack', 
                 load_mst = False, 
                 mst_path = '',
                 index_path = '',
                 min_box_size=1,
                 ) -> None:
        self.metric = metric
        self.approach = approach
        self.load_mst  = load_mst
        self.mst_path = mst_path
        self.index_path = index_path
        self.min_box_size = min_box_size

        if metric!='euclidean' and approach=='mlpack':
            self.approach = 'nx'

        # Parameters to set if they were precomputed
        self.index = None
        self.mst = None
        self.sorted_edges = None
    
    def _merge_components_boxes(self, c_a, c_b, i_a, i_b, d, i):
        # If a has a box
        if not self.points_component[i_a] is None:

            # If b does not have a box -> Join b to a's box
            if self.points_component[i_b] is None:
                a_box_id = self.points_component[i_a]
                self.points_component[list(c_b)] = a_box_id
                self.components_info[a_box_id]['points'].extend(list(c_b))
                self.components_info[a_box_id]['size'] += len(c_b)
                self.components_info[a_box_id]['persistence'] = d
                self.components_info[a_box_id]['children'] += 1

            # If b has a box -> Create parent box
            else:
                parent_box_id = len(self.components_info)
                self.components_info.append({})
                self.components_info[parent_box_id]['id'] = parent_box_id
                self.components_info[parent_box_id]['size'] = len(c_a)+len(c_b)
                self.components_info[parent_box_id]['points'] = c_a
                self.components_info[parent_box_id]['points'].extend(c_b)
                self.components_info[parent_box_id]['persistence'] = d
                self.components_info[parent_box_id]['created_at'] = d
                self.components_info[parent_box_id]['children'] = 2

                a_box_id = self.points_component[i_a]
                b_box_id = self.points_component[i_b]
                self.components_info[a_box_id]['parent'] = parent_box_id
                self.components_info[a_box_id]['died_at'] = d
                self.components_info[a_box_id]['persistence'] = d-self.components_info[a_box_id]['created_at']
                self.components_info[b_box_id]['parent'] = parent_box_id
                self.components_info[b_box_id]['died_at'] = d
                self.components_info[b_box_id]['persistence'] = d-self.components_info[b_box_id]['created_at']

                persistence_density_a = self.components_info[a_box_id]['size']/d
                self.components_info[a_box_id]['persistence_density'] = persistence_density_a
                persistence_density_b = self.components_info[b_box_id]['size']/d
                self.components_info[b_box_id]['persistence_density'] = persistence_density_b

                self.points_component[list(c_a)] = parent_box_id
                self.points_component[list(c_b)] = parent_box_id
                
        # If a does not have a box
        else:
            # If b has a box -> Join a to b's box
            if not self.points_component[i_b] is None:
                b_box_id = self.points_component[i_b]
                self.points_component[list(c_a)] = b_box_id
                self.components_info[b_box_id]['points'].extend(list(c_a))
                self.components_info[b_box_id]['size'] += len(c_a)
                self.components_info[b_box_id]['persistence'] = d
                self.components_info[b_box_id]['children'] += 1

            # If none has box
            else:
                # If a and b merged is bigger than min -> Create box
                if len(c_a)+len(c_b) >= self.min_box_size:
                    new_box_id = len(self.components_info)
                    self.components_info.append({})
                    self.components_info[new_box_id]['id'] = new_box_id
                    self.components_info[new_box_id]['size'] = len(c_a)+len(c_b)
                    self.components_info[new_box_id]['points'] = c_a
                    self.components_info[new_box_id]['points'].extend(c_b)
                    self.components_info[new_box_id]['persistence'] = d
                    self.components_info[new_box_id]['created_at'] = d
                    self.components_info[new_box_id]['children'] = 0

                    self.points_component[list(c_a)] = new_box_id
                    self.points_component[list(c_b)] = new_box_id

    def _get_components(self):

        for i in range(len(self.sorted_edges)):
            # Get points from the edge
            i_a, i_b = self.sorted_edges[i][0], self.sorted_edges[i][1]

            # Distance between points
            d = self.sorted_edges[i][2]

            # Get components the points belong to
            root_a, root_b = self.uf.find(i_a), self.uf.find(i_b)
            c_a = self.uf.component[root_a].points_ids.copy()
            c_b = self.uf.component[root_b].points_ids.copy()

            # Merge components 
            self._merge_components_boxes(c_a, c_b, i_a, i_b, d, i)
            self.uf.union(i_a, i_b)

        return self.components_info
    
    def fit(self, X:np.ndarray):
        self.uf = UnionFind(len(X), compute_hulls=False)
        self.points_component = np.array([None for i in range(len(X))])
        self.components_info = []

        if self.mst is None:
            self.mst = self._compute_mst(X)

        if self.sorted_edges is None:
            self.sorted_edges = self._compute_ordered_edges(self.mst)

        return self._get_components()
    


import os, errno
import networkx as nx
from .TopoMap import TopoMap
from .dataset_utils import fvecs_read, Index
from .utils import compute_mst_mlpack, compute_mst_ann
import numpy as np



def build_graph_mst(MST):
    '''
    Build nx graph from emst
    '''
    G = nx.Graph()
    n_nodes = len(MST) + 1
    G.add_nodes_from(range(n_nodes))

    edges = [0]*len(MST)
    for i in range(len(MST)):
        edges[i] = (int(MST[i,0]),int(MST[i,1]),MST[i,2])
    G.add_weighted_edges_from(edges)

    return G

def compute_difference_graphs(G1,G2):
    '''
    Compute statistics about two graphs MST graphs G1 and G2
    '''
    n_equal_edges = 0
    w_equal_edges = 0
    for edge in G1.edges.data("weight"):
        if(edge[1] in G2[edge[0]]):
            n_equal_edges+=1
            w_equal_edges += edge[2]
    return n_equal_edges, w_equal_edges

def compute_persitance_mst(mst):
    n_edges = len(mst)
    persistence = [(0,mst[i,2]) for i in range(n_edges)]
    return persistence


def plot_persitance_mst_and_dist(mst_ANN, mst_MLpack,dataset_name,):
    import gudhi
    import matplotlib.pyplot as plt
    max_weight_emst = np.max(mst_MLpack[:,2])
    max_weight_amst = np.max(mst_ANN[:,2])
    
    mst_MLpack[:,2] = mst_MLpack[:,2]/ max_weight_emst
    mst_ANN[:,2] = mst_ANN[:,2] / max_weight_amst
    
    I_mlpack = compute_persitance_mst(mst_MLpack)
    I_ANN = compute_persitance_mst(mst_ANN)

    dist = gudhi.bottleneck_distance(I_ANN, I_mlpack)


    fig, axs = plt.subplots(1, 2, figsize=(12, 4))
    
    gudhi.plot_persistence_diagram(I_mlpack,axes = axs[0])
    axs[0].set_title('Persistance - EMST')

    gudhi.plot_persistence_diagram(I_ANN, axes = axs[1])

    axs[1].set_title('Persistance - AMST')

    
    fig.suptitle(f'Persistence diagrams of MSTs - {dataset_name} dataset')
    fig.tight_layout()
    plt.show()
    

    return dist


def Compare_MSTs(EMST, ANN_MST,dataset_name):
    mst_mlpack_nx = build_graph_mst(EMST)
    mst_ANN_nx = build_graph_mst(ANN_MST)
    n_equal_edges, w_equal_edges =compute_difference_graphs(mst_ANN_nx,mst_mlpack_nx)
    n_edges = len(EMST)
    #ordering to avoid float point precision
    EMST = EMST[EMST[:, 2].argsort()]
    ANN_MST = ANN_MST[ANN_MST[:, 2].argsort()]
    total_weight = EMST[:,2].sum()
    approx_weight = ANN_MST[:,2].sum()
    percentual_weight_error = (approx_weight - total_weight) /total_weight 

    bn_dist = plot_persitance_mst_and_dist(ANN_MST, EMST,dataset_name)

    return {'Bottleneck distance':bn_dist,'# equal edges':n_equal_edges,'percentual equal edges':n_equal_edges/n_edges,'weight equal edges':w_equal_edges,'percentual weight of equal edges':w_equal_edges/total_weight, "Total weight error":percentual_weight_error}

def MST_test_topology(data_path, Index_path,dataset_name, save_folder,save_msts = False ):
    



    print("--")
    points  = fvecs_read(data_path)
    index = Index(Index_path)
    mst_ANN, ANN_building_time = compute_mst_ann(points,index, compute_time= True)
     
    mst_ANN_nx = build_graph_mst(mst_ANN)
    
    
    
    mst_mlpack, mlpack_building_time =  compute_mst_mlpack(points, compute_time = True)
    
    mst_mlpack_nx = build_graph_mst(mst_mlpack)

    bn_dist = plot_persitance_mst_and_dist(mst_ANN, mst_mlpack,dataset_name)
    n_equal_edges, w_equal_edges =compute_difference_graphs(mst_ANN_nx,mst_mlpack_nx)
    n_edges = len(mst_ANN)
    total_weight = mst_mlpack[:,2].sum()
    approx_weight = mst_ANN[:,2].sum()
    percentual_weight_error = abs(approx_weight/total_weight -1)

    if(save_msts):
        mst_filename = f'{save_folder}/mst_{dataset_name}.npy'
        vamana_mst_filename = f'{save_folder}/Vamana_mst_{dataset_name}.npy'
        # silentremove(mst_filename)
        # silentremove(vamana_mst_filename)
        np.save(mst_filename,mst_mlpack, allow_pickle=True)
        np.save(vamana_mst_filename,mst_ANN, allow_pickle= True)
        
    

    return [mlpack_building_time,ANN_building_time,bn_dist,n_equal_edges,n_equal_edges/n_edges,w_equal_edges,w_equal_edges/total_weight, percentual_weight_error]




        




import numpy as np

from .utils import get_hull

import plotly.express as px
from plotly.subplots import make_subplots
import plotly.graph_objects as go
from plotly.express.colors import sample_colorscale

def plot_topomap_comparison_highlight(proj_original, proj_new, 
                                      components_to_highligth, df_comp,
                                      hiertopomap=None,
                                      colors=[]
                                      ):
    

    highligth = np.zeros(shape=proj_original.shape[0])

    for i, comp in enumerate(components_to_highligth):
        highligth[df_comp.loc[comp]['points']] = i+1
    
    fig = make_subplots(rows=1, cols=2,
                        specs=[[{'type': 'xy'},
                                {'type': 'xy'}]],
                        subplot_titles=('Original Projection - TopoMap',
                                        'New Projection - Hierarchical TopoMap'),
                        horizontal_spacing = 0.02)
    
    highligth_values = np.unique(highligth)
    l = len(np.unique(highligth))
    if not 0 in highligth_values:
        l += 1

    if len(colors)==0:
        colors = px.colors.qualitative.T10

    num_colors = (len(components_to_highligth)//10)+1
    colors = num_colors*colors

    for i in range(l):
        if i==0:
            if not 0 in highligth_values:
                continue
            name = 'Other points'
        else:
            name = f'Component {components_to_highligth[i-1]}'

        fig.add_trace(
            go.Scatter(x=proj_original[highligth==i,0], 
                    y=proj_original[highligth==i,1],
                    mode='markers',
                    #opacity=0.5,
                    marker=dict(
                        color=colors[i],
                        size=2,
                    ),
                    name=str(name),
                    legendgroup=str(name),
                    ),
            row=1, col=1
        )

    if not hiertopomap is None:
        alphas = []
        for c in hiertopomap.components_to_scale:
            alphas.append(hiertopomap.components_info[c]['alpha'])

        #min_alpha = min(alphas)
        #range_alpha = max(alphas)-min_alpha
        max_alpha = max(alphas)
        max_color = px.colors.sample_colorscale('Greys', [1])[0]

        for j in range(len(hiertopomap.components_info)):
            comp_ids = hiertopomap.components_info[j]['points']
            if 'hull' in hiertopomap.components_info[j].keys():
                hull = hiertopomap.components_info[j]['hull']
                points_ids = [comp_ids[i] for i in hull.vertices]
                points = list(hiertopomap.projections[points_ids,:])
                points.append(points[0])
                xs, ys = zip(*points)

                #alpha_scaled = (hiertopomap.components_info[j]['alpha']-min_alpha)/range_alpha
                alpha_scaled = (hiertopomap.components_info[j]['alpha']-1)/(1.1*max_alpha-1)
                hull_color = sample_colorscale('Greys', [alpha_scaled])[0]

                fig.add_trace(go.Scatter(x=xs, y=ys,
                                fill='toself', 
                                fillcolor = hull_color,
                                line_color=max_color,
                                opacity=0.5,
                                line_width=1,
                                text=f'Component {j}',
                                name='Components', legendgroup='Components',
                                showlegend=False,
                                ),
                            row=1, col=2)
                
    for i in range(l):
        if i==0:
            if not 0 in highligth_values:
                continue
            name = 'Other points'
        else:
            name = f'Component {components_to_highligth[i-1]}'

        fig.add_trace(
            go.Scatter(x=proj_new[highligth==i,0], 
                    y=proj_new[highligth==i,1],
                    mode='markers',
                    marker=dict(
                        color=colors[i],
                        size=2,
                    ),
                    name=name,
                    legendgroup=name,
                    showlegend=False
                    ),
            row=1, col=2
        )

    fig.update_layout(margin = dict(t=75, l=25, r=25, b=25),
            height=500,
            width=1000,
            legend= {'itemsizing': 'constant',
                     'title': 'Components'},
            xaxis=dict(showticklabels=False,
                       showline=True, linewidth=1, linecolor='black', mirror=True), 
            yaxis=dict(showticklabels=False,
                       showline=True, linewidth=1, linecolor='black', mirror=True),
            xaxis2=dict(showticklabels=False,
                        showline=True, linewidth=1, linecolor='black', mirror=True), 
            yaxis2=dict(showticklabels=False,
                        showline=True, linewidth=1, linecolor='black', mirror=True),
            plot_bgcolor = "white"
            )

    return fig

def plot_projections_discrete_feature(projections,
                                    df_data,
                                    column_color,
                                    column_values=[],
                                    colors=[],
                                    hiertopomap=None,
                                    legend_title='',
                                    low_opacity=False,
                                    show_hulls=True,
                                    topomap=False
                                    ):
    if len(column_values) == 0:
        column_values = list(df_data[column_color].unique())
    if len(colors)==0:
        colors = px.colors.qualitative.T10
    if legend_title=='':
        legend_title=column_color

    fig = go.Figure()

    if show_hulls:
        for c in hiertopomap.components_to_scale:
            comp_ids = hiertopomap.components_info[c]['points']

            if topomap:
                hull = hiertopomap.components_info[c]['hull']
                points = projections[comp_ids]
                hull = get_hull(points)
                points_hull = [points[i] for i in hull.vertices]
                points_hull.append(points_hull[0])
                xs, ys = zip(*points_hull)

            else:
                hull = hiertopomap.components_info[c]['hull']
                points_ids = [comp_ids[i] for i in hull.vertices]
                points = list(hiertopomap.projections[points_ids,:])
                points.append(points[0])
                xs, ys = zip(*points)

            fig.add_trace(go.Scatter(x=xs, y=ys,
                            fill='toself', 
                            fillcolor = '#CCCCCC',
                            line_color='#808080',
                            opacity=0.5,
                            line_width=1,
                            text=f'Component {c}',
                            name='Components', legendgroup='Components',
                            showlegend=False,
                            marker=dict(size=1)
                            )
                        )

    if low_opacity:
        opacity_points = 0.5
    else:
        opacity_points = 1

    for i,c in enumerate(column_values):
        points_id = df_data[df_data[column_color]==c].index.to_list()
        fig.add_trace(go.Scatter(x=projections[points_id,0], 
                                y=projections[points_id,1],
                                customdata=points_id,
                                mode='markers',
                                opacity=opacity_points,
                                marker=dict(
                                    color=colors[i],
                                    size=2,
                                ),
                                name=str(c),
                                showlegend=True
                                )
            )
        
    fig.update_layout(margin = dict(t=25, l=25, r=25, b=25),
                        height=500,
                        width=550,
                        legend= {'itemsizing': 'constant',
                        'title': legend_title},
                        plot_bgcolor = "white",
                        xaxis=dict(showticklabels=False,
                        showline=True, linewidth=1, linecolor='black', mirror=True), 
                        yaxis=dict(showticklabels=False,
                        showline=True, linewidth=1, linecolor='black', mirror=True),
                    )
    
    return fig

def plot_hierarchical_treemap(df_comp, color='died_at'):
    fig = go.Figure(go.Treemap(
            labels=df_comp['id'],
            parents=df_comp['parent'],
            values=df_comp['size'],
            branchvalues='total',
            marker=dict(
                colors=df_comp[color],
                colorscale='Teal',
                showscale=True,
                colorbar=dict(
                    title='Persistence',
                    titleside='right',
                    thickness=15,
                    x=1
                )),
            hovertemplate='<b>Component #%{label} </b> <br> Points: %{value}<br> Persistence: %{color:.2f}<br> Parent: #%{parent}',
            tiling=dict(packing='squarify')
            ),
        )

    fig.update_layout(margin = dict(t=50, l=25, r=25, b=25),
                    title='TopoTree',
                    height=500,
                    width=800
                    )
    
    return fig

def plot_icicle(df_comp, color='died_at', color_title='Persistence'):
    fig = go.Figure(go.Icicle(
            labels=df_comp['id'],
            parents=df_comp['parent'],
            values=df_comp['size'],
            marker=dict(
                colors=df_comp[color],
                colorscale='Teal',
                showscale=True,
                colorbar=dict(
                    title=color_title,
                    titleside='right',
                    thickness=15,
                    x=1
                )),
            root_color="lightgrey",
            hovertemplate='<b>Component #%{label} </b> <br> Points: %{value}<br> Persistence: %{color:.2f}<br> Parent: #%{parent}',
            tiling = dict(orientation='v')
            ),
        )

    fig.update_layout(margin = dict(t=50, l=25, r=25, b=25),
                    title='Icicle',
                    height=500,
                    width=800
                    )
    
    return fig



import numpy as np 
from sklearn.base import BaseEstimator, ClassNamePrefixFeaturesOutMixin, TransformerMixin
from .utils import closest_edge_point_simplices, find_angle, Transform, fix_rotation, build_mst_ann, compute_mst_mlpack, compute_mst_nx

from .UnionFindComponents import UnionFind

class TopoMap(BaseEstimator, ClassNamePrefixFeaturesOutMixin, TransformerMixin):
    """Generate TopoMap projection of the input data points.

    Parameters
    ----------

    metric : {'cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan',' braycurtis', 'canberra', \
                'chebyshev', 'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis', \
                'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath',\
                'sqeuclidean', 'yule'} or a callable, default='euclidean'
        Distance metric to compute the MST. Check out sklearn's pairwise_distances to see other options.

    approach: {'mlpack', 'nx', 'ANN'}, default='mlpack'
        Approach for computing the MST. If 'mlpack', mlpack's function emst is used. In this case, the only 
        metric available is euclidean. If 'nx', it uses sklearn's pairwise_distance function followed by 
        networkx's minimum_spanning_tree function. If 'ANN', it first compute a  DiskANN's 
        nearest neighbor memory index and then compute the MST over this index. In  this case, only "euclidean", 
        "l2", "mips" and "cosine" are available

    load_mst: Bool, default=False
        Option to load precomputed mst. If set to True, you need to provide the mst_path.

    mst_path: string, default=''
        Path to the precomputed mst. Only used if load_mst is True.

    The following parameters are only used if approach == 'ANN'

    - **index_directory**: directory that will be used to store the index and temporarily store intermediate files
    - **vector_dtype**: The "points" datatype. If not passed,  we use the `data.dtype` if np array. 
    | Metric \ Datatype | np.float32 | np.uint8 | np.int8 |
    |-------------------|------------|----------|---------|
    | L2 | Euclidean    |      ✅     |     ✅    |    ✅    |
    | MIPS              |      ✅     |     ❌    |    ❌    |
    | Cosine            |      ✅     |     ✅    |    ✅    |
    
    - **graph_degree**: The degree of the graph index, typically between 60 and 150. A larger maximum degree will
      result in larger indices and longer indexing times, but better search quality.   
    - **complexity**: The size of the candidate nearest neighbor list to use when building the index. Values between 75
      and 200 are typical. Larger values will take more time to build but result in indices that provide higher recall
      for the same search complexity. Use a value that is at least as large as `graph_degree` unless you are prepared
      to compromise on quality
    - **num_threads**: Number of threads to use when creating this index. `0` is used to indicate all available
      logical processors should be used.
    - **alpha**: The alpha parameter (>=1) is used to control the nature and number of points that are added to the
      graph. A higher alpha value (e.g., 1.4) will result in fewer hops (and IOs) to convergence, but probably more
      distance comparisons compared to a lower alpha value.
    - **index_prefix**: The prefix of the index files. Defaults to "ann".
    - **tmp_foldeer**:  If True, deletes the directory where the index was built.
        
    """

    def __init__(self, 
                 metric='euclidean', 
                 approach = 'mlpack', 
                 load_mst = False, 
                 mst_path = '',
                 index_directory:str = 'tmp/',
                vector_dtype = np.float32,
                graph_degree:int = 64,
                complexity:int =100,
                num_threads:int=0,
                alpha:float = 1.2,
                index_prefix:str = 'ann'
                 ) -> None:
        self.metric = metric
        self.approach = approach
        self.load_mst  = load_mst
        self.mst_path = mst_path
        self.index_directory = index_directory
        self.vector_dtype = vector_dtype
        self.graph_degree = graph_degree
        self.complexity = complexity
        self.num_threads = num_threads
        self.alpha = alpha
        self.index_prefix = index_prefix

        if metric!='euclidean' and approach=='mlpack':
            self.approach = 'nx'

        # Parameters to set if they were precomputed
        self.mst = None
        self.sorted_edges = None

    def _compute_mst(self, X):
        if self.load_mst:
            mst = np.load(self.mst_path,allow_pickle= True)
        elif self.approach == 'mlpack':
            mst = compute_mst_mlpack(X)
        elif self.approach == 'nx':
            mst = compute_mst_nx(X, metric=self.metric)
        elif self.approach == 'ANN':
            metric = self.metric
            if(metric == 'euclidean'):
                metric = 'l2'
            elif(metric not in ['l2','mips', 'cosine']):
                print("only euclidean, mips, cosine and l2 metrics are supported for ANN approach")
            #Build VamanaIndex and mst
            mst = build_mst_ann(points = X,
                                index_directory = self.index_directory,
                                vector_dtype = self.vector_dtype,
                                distance_metric = metric,
                                graph_degree = self.graph_degree,
                                complexity = self.complexity,
                                num_threads = self.num_threads,
                                alpha = self.alpha,
                                index_prefix = self.index_prefix 
                                )
        else:
            print("Approach isn't supported, please choose between ANN, mlpack and nx")
        return mst
      
    def _compute_ordered_edges(self, mst):
        sorted_edges = mst[mst[:, 2].argsort()]
        return sorted_edges
    
        
    def _get_mst(self):
        return self.mst
        
    def _get_sorted_edges(self):
        return self.sorted_edges

    def _rotate_component(self, 
                         component_ids:list,
                         ref_point:np.ndarray, 
                         component_simplices:list,
                         direction='top') -> np.ndarray:

        component_points = self.projections[component_ids,:]

        if len(component_simplices)==1:
            closest_edge = self.projections[component_simplices[0]]
            edge_i = component_simplices[0]
        else:
            closest_edge, edge_i = closest_edge_point_simplices(component_simplices,
                                                        self.projections,
                                                        ref_point)
            
        t = Transform()
        t.cos, t.sin = find_angle(closest_edge)
        component_points = t.rotate(component_points)

        id_0 = component_ids.index(edge_i[0])
        id_1 = component_ids.index(edge_i[1])
        closest_edge = np.array([component_points[id_0,:],
                                    component_points[id_1,:]])
        component_points = fix_rotation(closest_edge, 
                                        component_points, 
                                        direction=direction)

        return component_points, [id_0,id_1]
    
    def _translate_component(self, 
                            component_points:list, 
                            edge_i:np.ndarray,
                            to_point:list) -> np.ndarray:
        
        if component_points[edge_i[0], 0] <= component_points[edge_i[1], 0]:
            t = Transform(x = to_point[0]-component_points[edge_i[0], 0], 
                          y = to_point[1]-component_points[edge_i[0], 1])
            component_points = t.translate(component_points)

        else:
            t = Transform(x = to_point[0]-component_points[edge_i[1], 0], 
                          y = to_point[1]-component_points[edge_i[1], 1])
            component_points = t.translate(component_points)

        return component_points

    def _project_points(self):
        
        for i in range(len(self.sorted_edges)):
            # Get points from the edge
            i_a, i_b = self.sorted_edges[i][0], self.sorted_edges[i][1]
            p_a, p_b = self.projections[i_a,:], self.projections[i_b,:]

            # Distance between points
            d = self.sorted_edges[i][2]
            
            # Get components the points belong to
            root_a, root_b = self.uf.find(i_a), self.uf.find(i_b)
            c_a = self.uf.component[root_a].points_ids.copy()
            c_b = self.uf.component[root_b].points_ids.copy()

            proj_c_a, proj_c_b = self.projections[list(c_a)], self.projections[list(c_b)]

            simplices_a = self.uf.component[root_a].hull_simplices
            simplices_b = self.uf.component[root_b].hull_simplices

            # Rotate the first to be the topmost
            proj_c_a, edge_t = self._rotate_component(c_a, p_a, simplices_a, direction='top')
            # Rotate the second to be the bottomost
            proj_c_b, edge_b = self._rotate_component(c_b, p_b, simplices_b, direction='bottom')

            # Translate components
            proj_c_a = self._translate_component(proj_c_a, edge_t, to_point=[0,0])
            proj_c_b = self._translate_component(proj_c_b, edge_b, to_point=[0,d])

            self.projections[list(c_a), :] = proj_c_a
            self.projections[list(c_b), :] = proj_c_b

            # Merge components 
            self.uf.union(i_a, i_b, self.projections)

        return self.projections
    
    def fit_transform(self, X:np.ndarray):
        """Fit X into an embedded space and return that transformed output.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)

        Returns
        -------
        X_new : ndarray of shape (n_samples, n_components)
            Embedding of the training data in two-dimensional space.
        """

        self.projections = np.zeros(shape=(len(X), 2), dtype=np.float32)
        self.uf = UnionFind(len(X))

        if self.mst is None:
            self.mst = self._compute_mst(X)

        if self.sorted_edges is None:
            self.sorted_edges = self._compute_ordered_edges(self.mst)

        self.projections = self._project_points()

        return self.projections



import numpy as np
from scipy.sparse import csr_matrix
import os
def ivecs_write(fname, m):
    n, d = m.shape
    m1 = np.empty((n, d + 1), dtype='int32')
    m1[:, 0] = d
    m1[:, 1:] = m
    m1.tofile(fname)


def fvecs_write(fname, m):
    m = m.astype('float32')
    ivecs_write(fname, m.view('int32'))



def fvecs_read(filename, c_contiguous=True):
    fv = np.fromfile(filename, dtype=np.float32)
    if fv.size == 0:
        return np.zeros((0, 0))
    dim = fv.view(np.int32)[0]
    assert dim > 0
    fv = fv.reshape(-1, 1 + dim)
    if not all(fv.view(np.int32)[:, 0] == dim):
        raise IOError("Non-uniform vector sizes in " + filename)
    fv = fv[:, 1:]
    if c_contiguous:
        fv = fv.copy()
    return fv

def read_and_convert(file,nbits):
    '''Receive the file and nbits and return the integer that is encoded.
    nbits = 8 or 4'''
    encoded_int = file.read(nbits)
    encoded_int = int.from_bytes(encoded_int,byteorder='little',signed = False)
    return encoded_int


def Index_read(filename, verbose = False):
        #I'm not using the start since we won't search, but if necessary it can be returned from this function
        file = open(filename, "rb")

        expected_file_size = read_and_convert(file,8)
        max_observed_degree = read_and_convert(file,4)
        start = read_and_convert(file,4)
        file_fronzen_pts = read_and_convert(file,8)       
        bytes_read = 8 + 4 + 4 + 8 
        nodes_read = 0
        #I don't know the length of the graph in advance, so I'm using a dict instead of appending several times in a list of lists
        adj_dict ={}
        cc =0
        while( bytes_read != expected_file_size):  
            number_of_neighbors = read_and_convert(file,4)
            cc+=number_of_neighbors
            if( number_of_neighbors == 0):
                print(f"ERROR: Point Found with no out-neighbours, point# {nodes_read} \n")        
            buffer = file.read(4* number_of_neighbors)
            ngb = np.frombuffer(buffer, dtype=np.int32)
            adj_dict[nodes_read] = ngb
            nodes_read+=1
            bytes_read += 4*(number_of_neighbors + 1)
        file.close()
        if(verbose):
            print(f'From graph header, expected file size:{expected_file_size}\n max_observed_degree:{max_observed_degree} \n file_frozen_pts: {file_fronzen_pts}\n ')
            print(f'done. Index has {nodes_read} nodes')
        return adj_dict

def turn_symmetric(adj_dict):
    '''
    Assumes that adj_dict is a dictionary that has each adjacency_list associated with each key
    Returns a adj_dict where 
    '''

    for k in adj_dict.keys():
        for ngb in adj_dict[k]:
            if k not in adj_dict[ngb]:
                adj_dict[ngb] = np.append(adj_dict[ngb],k)

    return adj_dict

def Compute_distance(dataset,i,j, aproximate_almost_zero = False,seed = None):
    from random import random
    distance = np.linalg.norm(dataset[i,:]- dataset[j,:])
    if (aproximate_almost_zero):
        if(seed):
            random.seed(seed)
        random_noise = 1 +random()
        distance = distance + 1e-7*random_noise
    return distance

def Compute_adj_matrix(adj_dict, input = None, input_file ="", turn_sym = False, aproximate_almost_zero = False, ):
    '''
    Compute adjacency matrix using sparse scipy representation. 
    Assumes that adj_dict is a dictionary that has each adjacency_list associated with each key
    input_file is the path for the input_file. Assumed to be on fvecs format for a while
    '''
    if(input_file):
        input = fvecs_read(input_file)
    number_of_nodes = len(adj_dict)

    
    if(turn_sym):
        adj_dict = turn_symmetric(adj_dict)
    total_edges = 0
    for v in adj_dict.values():
        total_edges += len(v)
    distances = np.zeros(total_edges)
    indptr = np.zeros(number_of_nodes+1)
    edges_read = 0
    indices = np.zeros(total_edges)
    for i in range(number_of_nodes):
        neighbors = adj_dict[i]
        for neighbor in neighbors:
            distances[edges_read] = Compute_distance(input,i,neighbor,aproximate_almost_zero)
            indices[edges_read] = neighbor
            edges_read+=1
        
        indptr[i+1] = indptr[i]+len(neighbors)
    
    
    sparse_matrix = csr_matrix((distances,indices,indptr),shape=(number_of_nodes,number_of_nodes))
        
    return sparse_matrix
     


def Compute_MST_from_adj(adj_matrix):
    from scipy.sparse.csgraph import  minimum_spanning_tree
    from scipy.sparse import issparse
    is_sparse = issparse(adj_matrix)
    if(is_sparse):
        adj_matrix.data += 1
        MST = minimum_spanning_tree(adj_matrix)
        MST.data -= 1
    else:
        MST = minimum_spanning_tree(adj_matrix)
    return MST


class Index:
    """Read Index generated using DISKANN microsoft implementation
    """
    def __init__(self, index_path) -> None:
        self.index_path = index_path

    def get_mst(self, points):
        Index = Index_read(self.index_path)
        sparse_adj_matrix = Compute_adj_matrix(Index, input = points)
        mst = Compute_MST_from_adj(sparse_adj_matrix)
        return mst

def silentremove(filename):
    '''
    Delete a file if it exists without raise an error.  
    '''
    try:
        os.remove(filename)
    except: 
        pass 



    

    



import numpy as np 
from numpy.linalg import norm
from scipy.spatial import ConvexHull
from pathlib import Path
import shutil
import os
import mlpack
import diskannpy as dap

from .dataset_utils import Index, silentremove

class AuxHull():
    def __init__(self, points, simplices):
        self.points = points
        self.simplices = simplices
        self.vertices = np.array(list(range(len(points))))

def get_hull(points:np.ndarray, aligned_points = False):
    if len(points) == 1:
        hull = AuxHull(points=points,
                       simplices=np.array([[0,0]]))
    elif len(points) == 2:
        hull = AuxHull(points=points,
                       simplices=np.array([[0,1], [1,0]]))
    elif aligned_points:
        simplices = [[i,(i+1)%len(points)] for i in range(len(points))]
        hull = AuxHull(points = points,
                       simplices = np.array(simplices))
    else:
        hull = ConvexHull(points)
        
    return hull

def distance_segment_point(segment, p):
    a, b = segment
    
    if a[0]==b[0] and a[1]==b[1]:
        return norm(a-p)
    if (a[0]==p[0] and a[1]==p[1]) or (p[0]==b[0] and p[1]==b[1]):
        return 0
    else:
        proj_p_ab = np.dot((p-a), (b-a))
        norm_proj_p_ab = proj_p_ab / (norm(b-a)**2)

        if norm_proj_p_ab <= 0:
            closest_point = a
        elif norm_proj_p_ab>= 1:
            closest_point = b
        else:
            closest_point = a + norm_proj_p_ab*(b-a)

        return norm(closest_point-p)

def closest_edge_point(hull, 
                       ref_point:np.ndarray):
    '''
    Find edge of convex hull that is closer to the reference point.
    Returns the points of the edge and its indexes.
    '''
    closest_i = -1
    closest_dist = np.inf

    for i in range(len(hull.simplices)):
        a, b = hull.points[hull.simplices[i]]
        d = distance_segment_point((a, b), ref_point)
        
        if d < closest_dist:
            closest_dist = d
            closest_i = i

    closest_edge = hull.points[hull.simplices[closest_i]]

    return closest_edge, hull.simplices[closest_i]

def closest_edge_point_simplices(component_simplices, 
                                 projections,
                                 ref_point:np.ndarray):
    '''
    Find edge of convex hull that is closer to the reference point.
    Returns the points of the edge and its indexes.
    '''
    closest_i = -1
    closest_dist = np.inf

    for i in range(len(component_simplices)):
        a, b = projections[component_simplices[i],:]
        d = distance_segment_point((a, b), ref_point)
        
        if d < closest_dist:
            closest_dist = d
            closest_i = i

    closest_simplice = component_simplices[closest_i]
    closest_edge = projections[closest_simplice,:]

    return closest_edge, closest_simplice

def find_angle(segment:np.ndarray):
    '''
    Find angle to rotate component so that segment is paralel to the vertical axis
    '''
    p1, p2 = segment
    
    if p1[1] == p2[1]:
        # Angle is 0 degree
        return 1, 0
    if p1[0] == p2[0]:
        # Angle is 90 degree
        return 0, 1

    # Use rightmost vertice as reference for angle
    if p1[0] < p2[0]:
        vec = p2 - p1
    else:
        vec = p1 - p2

    hip = norm(vec)
    vec[0] /= hip
    vec[1] /= hip

    cos = vec[0]
    sin = np.sqrt(1 - cos**2)
    if vec[1] >= 0:
        sin = -1*sin

    return cos, sin

class Transform():
    def __init__(self, x=0, y=0, sin=0, cos=1, scalar=1) -> None:
        self.x = x
        self.y = y
        self.sin = sin
        self.cos = cos
        self.scalar = scalar

    def translate(self, points:np.ndarray) -> np.ndarray:
        trans_points = np.zeros(shape=(len(points),2))
        trans_points = points + np.array([self.x, self.y])
        return trans_points

    def rotate(self, points:np.ndarray) -> np.ndarray:
        rot_points = np.zeros(shape=(len(points),2))
        rot_points[:,0] = points[:,0] * self.cos - points[:,1] * self.sin
        rot_points[:,1] = points[:,0] * self.sin + points[:,1] * self.cos
        return rot_points
    
    def scale(self, points:np.ndarray) -> np.ndarray:
        sca_points = self.scalar * points
        return sca_points

    def transform(self, points:np.ndarray) -> np.ndarray:
        transform_points = np.zeros(shape=(len(points),2))
        transform_points = self.translate(points)
        transform_points = self.rotate(transform_points)
        return transform_points
    
def fix_rotation(segment:np.ndarray, 
                 points:np.ndarray, 
                 direction='top') -> np.ndarray:
    '''
    Rotates points so that the segment is topmost of bottomost edge
    '''
    new_points = points.copy()

    if direction == 'top':
        max_y = points[:,1].max()
        if max_y > segment[:,1].max():
            # Rotate 180 degrees
            t = Transform(sin=0, cos=-1)
            new_points = t.rotate(new_points)

    elif direction == 'bottom':
        min_y = points[:,1].min()
        if min_y < segment[:,1].min():
            # Rotate 180 degrees
            t = Transform(sin=0, cos=-1)
            new_points = t.rotate(new_points)

    return new_points



def recompute_distance_MST(points, mst):
    n_edges = len(mst)
    for i in range(n_edges):
        e1 = mst[i,0]
        e2 = mst[i,1]
        distance = np.linalg.norm(points[e1,:] - points[e2,:])
        mst[i,2] = distance
    return mst

def compute_mst_mlpack(points, compute_time = False, recompute_distance = False):
    import time
    start = time.time()
    d = mlpack.emst(input_ = points)
    end = time.time()
    total_time = end- start
    d = d["output"]
    mst = np.array([[int(d[i,0]),int(d[i,1]),d[i,2]] for i in range(len(d))],dtype='O')
    if(recompute_distance):
        mst = recompute_distance_MST(points, mst)
    if (compute_time):
        output = (mst,total_time)
    else:
        output = mst
    return output

def compute_mst_ann(points, index, compute_time= False, recompute_distance = False):
    import time 
    start = time.time()
    mst = index.get_mst(points)
    end = time.time()
    total_time = end - start
    mst = mst.tocoo()
    rows = mst.row
    cols = mst.col
    data = mst.data
    mst = np.array([[int(rows[i]),int(cols[i]),data[i]] for i in range(len(rows))],dtype='O')

    if(recompute_distance):
        mst = recompute_distance_MST(points, mst)
    
    if (compute_time):
        output = (mst,total_time)
    else:
        output = mst
    return output

def compute_mst_nx(X, metric='cosine'):
    from sklearn.metrics.pairwise import pairwise_distances
    import networkx as nx

    n = len(X)
    D = pairwise_distances(X, X, metric=metric)

    G = nx.Graph()
    for i in range(n):
        for j in range(1,n):
            if i!=j:
                G.add_edge(i, j, weight=D[i,j])

    T = nx.minimum_spanning_tree(G)

    edges = [[e[0], e[1], e[2]['weight']] for e in T.edges(data=True)]
    edges = np.array(edges, dtype = object)

    return edges


def build_vamana_index(points: np.array,
                       index_directory:str = 'tmp/',
                       vector_dtype = np.float32,
                       distance_metric:str = 'l2',
                       graph_degree:int = 64,
                       complexity:int =100,
                       num_threads:int=0,
                       alpha:float = 1.2,
                       index_prefix:str = 'ann'
                       ):
    '''
    Generate a DiskANN index for a numpy array.

        
    ## Distance Metric and Vector Datatype Restrictions

    | Metric \ Datatype | np.float32 | np.uint8 | np.int8 |
    |-------------------|------------|----------|---------|
    | L2                |      ✅     |     ✅    |    ✅    |
    | MIPS              |      ✅     |     ❌    |    ❌    |
    | Cosine            |      ✅     |     ✅    |    ✅    |

    ### Parameters

    - **points*:  a numpy.ndarray of a supported dtype in 2 dimensions. 
    - **index_directory**: directory that will be used to store the index and temporarily store intermediate files
    - **vector_dtype**: The "points" datatype. If not passed,  we use the `data.dtype` if np array.
    - **distance_metric**: A `str`, strictly one of {"l2", "mips", "cosine"}. `l2` and `cosine` are supported for all 3
      vector dtypes, but `mips` is only available for single precision floats. 
    - **graph_degree**: The degree of the graph index, typically between 60 and 150. A larger maximum degree will
      result in larger indices and longer indexing times, but better search quality.   
    - **complexity**: The size of the candidate nearest neighbor list to use when building the index. Values between 75
      and 200 are typical. Larger values will take more time to build but result in indices that provide higher recall
      for the same search complexity. Use a value that is at least as large as `graph_degree` unless you are prepared
      to compromise on quality
    - **num_threads**: Number of threads to use when creating this index. `0` is used to indicate all available
      logical processors should be used.
    - **alpha**: The alpha parameter (>=1) is used to control the nature and number of points that are added to the
      graph. A higher alpha value (e.g., 1.4) will result in fewer hops (and IOs) to convergence, but probably more
      distance comparisons compared to a lower alpha value.
    - **index_prefix**: The prefix of the index files. Defaults to "ann".

    More information can be found on the library repository https://github.com/microsoft/DiskANN/tree/main/python
    '''
    dir_location = Path(index_directory)
    if(not dir_location.is_dir()):
        os.makedirs(index_directory)
    #remove files is they already exist
    silentremove(os.path.join(index_directory,index_prefix))
    silentremove(os.path.join(index_directory,index_prefix+'.data'))
    silentremove(os.path.join(index_directory,index_prefix+'_metadata.bin'))
    silentremove(os.path.join(index_directory,index_prefix+'_vectors.bin'))

    dap.build_memory_index(
                                data=points,
                                vector_dtype=vector_dtype,
                                distance_metric=distance_metric,
                                index_directory=index_directory,
                                graph_degree=graph_degree,
                                complexity=complexity,
                                num_threads=num_threads,
                                alpha = alpha,
                                index_prefix = index_prefix
                                )

    os.remove(os.path.join(index_directory,index_prefix+'.data'))
    os.remove(os.path.join(index_directory,index_prefix+'_metadata.bin'))
    os.remove(os.path.join(index_directory,index_prefix+'_vectors.bin'))




def build_mst_ann(points: np.array,
                index_directory:str = 'tmp/',
                vector_dtype = np.float32,
                distance_metric:str = 'l2',
                graph_degree:int = 64,
                complexity:int =100,
                num_threads:int=0,
                alpha:float = 1.2,
                index_prefix:str = 'ann',
                tmp_folder = True):
    '''
    Generate a aproximate MST by first computing a  DiskANN's 
    nearest neighbor memory index and then compute the MST over this index

    ### Parameters

    - **tmp_foldeer**:  If True, deletes the directory where the index was built.
    For a description of other parameters, please refer to the documentation of build_vamana_index function

    '''
    build_vamana_index(points = points,
                       index_directory = index_directory,
                       vector_dtype = vector_dtype,
                       distance_metric = distance_metric,
                       graph_degree = graph_degree,
                       complexity = complexity,
                       num_threads = num_threads,
                       alpha = alpha,
                       index_prefix = index_prefix)
    index = Index(os.path.join(index_directory,index_prefix))
    ANN_mst = compute_mst_ann(points,index)
    if(tmp_folder):
        shutil.rmtree(index_directory)
    return ANN_mst


import numpy as np
from .utils import get_hull, AuxHull

class Component:
    def __init__(self, 
                 point_id):
        self.points_ids = [point_id]
        self.hull_ids = [point_id]
        self.hull_simplices = [[point_id,point_id]]

class UnionFind:
    def __init__(self, n, 
                 compute_hulls=True):
        self.parent = [i for i in range(n)]
        self.size = [1]*n
        self.count = n
        self.component = self.make_components(n)
        self.compute_hulls = compute_hulls

    def make_components(self, n):
        component = []
        for i in range(n):
            c = Component(i)
            component.append(c)
        return component

    def find(self, node):
        while node != self.parent[node]:
            self.parent[node] = self.parent[self.parent[node]]
            node = self.parent[node]
        return node
    
    def find_component(self, node):
        node = self.find_id(node)
        return self.component[node]
    
    def compute_union_hull(self, root1, root2, projections):
        if self.size[root1]+self.size[root2] == 2:
            return [root1, root2]
        else:
            h1 = self.component[root1].hull_ids
            h2 = self.component[root2].hull_ids
            h1.extend(h2)

            if projections[h1,0].max() == projections[h1,0].min():
                min_id = projections[h1,1].argmin() 
                max_id = projections[h1,1].argmax() 
                return [h1[min_id], h1[max_id]]
            
            union_hull = get_hull(projections[h1,:])
            union_hull_ids = [h1[i] for i in union_hull.vertices]

            return union_hull_ids
    
    def union(self, node1, node2, projections=None):
        root1 = self.find(node1)
        root2 = self.find(node2)

        if root1 == root2:
            return

        if self.size[root1] >= self.size[root2]:
            self.parent[root2] = root1
            self.size[root1] += self.parent[root2]
            self.component[root1].points_ids.extend(self.component[root2].points_ids)

            if self.compute_hulls:
                new_hull_ids = self.compute_union_hull(root1, root2, projections)
                self.component[root1].hull_ids = new_hull_ids
                m = len(new_hull_ids)
                simplices = [(new_hull_ids[i], new_hull_ids[(i+1)%m]) for i in range(m)]
                self.component[root1].hull_simplices = simplices

            self.component[root2] = None
        else:
            self.parent[root1] = root2
            self.size[root2] += self.parent[root1]
            self.component[root2].points_ids.extend(self.component[root1].points_ids)

            if self.compute_hulls:
                new_hull_ids = self.compute_union_hull(root2, root1, projections)
                self.component[root2].hull_ids = new_hull_ids
                m = len(new_hull_ids)
                simplices = [(new_hull_ids[i], new_hull_ids[(i+1)%m]) for i in range(m)]
                self.component[root2].hull_simplices = simplices

            self.component[root1] = None
        
        self.count -= 1


from setuptools import setup, find_packages
from pathlib import Path

setup(
    name="topomap",
    version="0.0.1",
    packages=find_packages(),
    install_requires=[
        "matplotlib>=3.8.2",
        "mlpack>=4.3.0.post1",
        "networkx>=3.2.1",
        "numpy>=1.25.0",
        "pandas>=2.2.2",
        "plotly>=5.18.0",
        "scikit_learn>=1.4.1.post1",
        "scipy>=1.13.1",
        "diskannpy==0.7.0"
    ],
    author="Vitoria Guardieiro, Felipe Inagaki de Oliveira, Harish Doraiswamy, Luis Gustavo Nonato, Claudio Silva",
    author_email="vitoriaguardieiro@gmail.com",
    description="TopoMap++: A faster and more space efficient technique to compute projections with topological guarantees",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=[
        "Topological data analysis", 
        "Computational topology", 
        "High-dimensional data", 
        "Projection"
    ],
    url="https://github.com/viguardieiro/TopoMap",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



import pandas as pd
import numpy as np 

from .utils import Transform, get_hull
from .UnionFindComponents import UnionFind

from .TopoMap import TopoMap
from .TopoTree import TopoTree

class HierarchicalTopoMap(TopoMap):
    """Generate HierarchicalTopoMap projection of the input data points.

    Parameters
    ----------

    metric : {'cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan',' braycurtis', 'canberra', \
                'chebyshev', 'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis', \
                'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath',\
                'sqeuclidean', 'yule'} or a callable, default='euclidean'
        Distance metric to compute the MST. Check out sklearn's pairwise_distances to see other options.

    approach: {'mlpack', 'nx', 'ANN'}, default='mlpack'
        Approach for computing the MST. If 'mlpack', mlpack's function emst is used. In this case, the only 
        metric available is euclidean. If 'nx', it uses sklearn's pairwise_distance function followed by 
        networkx's minimum_spanning_tree function. If 'ANN', it uses DiskANN's precomputed approximate 
        nearest neighbor instead of computing a MST. In this case, you need to provide the index_path.

    load_mst: Bool, default=False
        Option to load precomputed mst. If set to True, you need to provide the mst_path.

    mst_path: string, default=''
        Path to the precomputed mst. Only used if load_mst is True.

    index_path: string, default=''
        Path to index file generated by DiskANN. Only used if approach is set to 'ANN'.

    min_points_component : int, default=2
        Minumum number of points for a component to have an assigned box.

    max_edge_length : float, default=-1 (ignore restriction)
        Optional parameter to control the maximum length for an edge to be processed.

    components_to_scale : list, default=[]
        List of component ids to be scaled. If it is empty, the method get_component_to_scale 
        will be used to select the components.

    max_scalar : float, default=20
        Maximum scalar alpha to scale a component.

    selection_method : {'k_components', 'min_persistence', 'min_size', 'max_persistence', \
                        'max_size'}, default='k_components'
        Method for selecting the components to be scale. Only used if components_to_scale is empty.

    k_components : int, default=5
        Number of components to select for scaling. The selection starts on the root of the tree
        and moves towards the leafs until we have selected k_components. If we reach a component
        wihtout childs, it is added to the list of selected components. Otherwise, we continue
        processing the component's childs until the end criteria is met.
        Only used if components_to_scale is empty and selection_method is 'k_components'.

    min_persistence_selection : float, default=0.1
        Minimum persistence a component must have to be selected. The selection starts on the root 
        of the tree and moves towards the leafs while the components have persistence above this value. 
        If we reach a component wihtout childs, it is added to the list of selected components. 
        If all childs of this component have persistence less then this value, the component
        is added to the list of selected components. Otherwise, we continue processing the component's 
        childs until there are no more components to explore.
        Only used if components_to_scale is empty and selection_method is 'min_persistence'.

    min_size_selection : float, default=0.1
        Minimum number of points a component must have to be selected. The selection starts on the root 
        of the tree and moves towards the leafs while the components have size above this value. 
        If we reach a component wihtout childs, it is added to the list of selected components. 
        If all childs of this component have size smaller then this value, the component is added to the 
        list of selected components. Otherwise, we continue processing the component's childs until 
        there are no more components to explore.
        Only used if components_to_scale is empty and selection_method is 'min_size'.

    max_persistence_selection : float, default=0.1
        Maximum persistence a component must have to be selected. The selection starts on the root 
        of the tree and moves towards the leafs until the components have persistent below this value. 
        If the component persistence below this value, it is added to the list of selected components. 
        Otherwise, we continue exploring the component's childs.
        Only used if components_to_scale is empty and selection_method is 'max_persistence'.

    max_size_selection : float, default=1
        Maximum number of points a component must have to be selected. The selection starts on the root 
        of the tree and moves towards the leafs until the components size is below this value. 
        If the component size below this value, it is added to the list of selected components. 
        Otherwise, we continue exploring the component's childs.
        Only used if components_to_scale is empty and selection_method is 'max_size'.

    verbose : bool, default=True
        Option to print information about the execution.
    """   

    def __init__(self,
                 metric='euclidean',
                 approach = 'mlpack', 
                 load_mst = False, 
                 mst_path = '',
                 index_path = '',
                 min_points_component=2,
                 max_edge_length=-1,
                 components_to_scale=[],
                 blowout_scalar=2,
                 max_scalar=20,
                 selection_method='k_components',
                 k_components=5,
                 min_persistence_selection=0.1,
                 min_size_selection=0,
                 max_persistence_selection=0.1,
                 max_size_selection=1,
                 verbose=True
                 ) -> None:

        # MST-related parameters
        self.verbose = verbose
        self.metric = metric
        self.approach = approach
        self.load_mst  = load_mst
        self.mst_path = mst_path
        self.index_path = index_path

        if metric!='euclidean' and approach=='mlpack':
            self.approach = 'nx'

        # Projection-related parameters
        self.min_points_component = min_points_component
        self.max_edge_length = max_edge_length
        self.components_to_scale = components_to_scale
        self.blowout_scalar = blowout_scalar
        self.max_scalar = max_scalar
        self.goal_length = max_edge_length

        # Options for get_component_to_scale
        self.selection_method = selection_method
        self.k_components = k_components
        self.min_persistence_selection = min_persistence_selection
        self.min_size_selection = min_size_selection
        self.max_persistence_selection = max_persistence_selection
        self.max_size_selection = max_size_selection

        # Parameters to set if they were precomputed
        self.index = None
        self.mst = None
        self.sorted_edges = None

    def _get_component_to_scale(self, 
                                X:np.ndarray,
                                df_comp=None, 
                                selection_method='k_components',
                                k_components=5,
                                min_persistence=0.1,
                                min_size=0,
                                max_persistence=0.1,
                                max_size=0):
        if df_comp is None:
            topotree = TopoTree(min_box_size=self.min_points_component)
            topotree.mst = self.mst
            topotree.sorted_edges = self.sorted_edges
            comp_info = topotree.fit(X)

            df_comp = pd.DataFrame.from_dict(comp_info)

        childs_map = {}
        for i in range(df_comp.shape[0]):
            childs_component = df_comp[df_comp['parent']==i].id.to_list()
            childs_map[i] = childs_component

        selected_comps = []

        if selection_method == 'k_components':
            ordered_comps = df_comp.sort_values(by=['size'], ascending=False).index.to_list()

            # Go down in the ordered list of components until we selected k
            i = 0
            while len(selected_comps) < k_components and i < len(ordered_comps):
                next_comp = ordered_comps[i]
                parent = df_comp.loc[next_comp,'parent']
                if parent in selected_comps:
                    selected_comps.remove(parent)
                selected_comps.append(next_comp)
                i += 1

            # Go up in the selected k and pick the parents that don't clash
            for i, c in enumerate(selected_comps):
                parents_clash = False
                parent_c = int(df_comp.loc[c,'parent'])
                while not parents_clash and not np.isnan(parent_c):
                    for j in range(len(selected_comps)):
                        if i==j:
                            continue
                        parent_j = df_comp.loc[selected_comps[j],'parent']
                        while not np.isnan(parent_j):
                            if parent_j == parent_c:
                                parents_clash = True
                                break
                            if not np.isnan(parent_j):
                                parent_j = df_comp.loc[parent_j,'parent']
                        if parents_clash:
                            break
                    if not parents_clash:
                        selected_comps = [parent_c if x==c else x for x in selected_comps]
                        c = parent_c
                        parent_c = int(df_comp.loc[c,'parent'])

        elif selection_method == 'min_persistence' or selection_method == 'min_size':
            if selection_method == 'min_persistence':
                feature = 'died_at'
                min_value = min_persistence
            else:
                feature = 'size'
                min_value = min_size

            parents = [df_comp.shape[0]-1]

            while len(parents) > 0:
                parents_aux = parents.copy()
                for parent in parents:
                    parents_aux.remove(parent)
                    if len(childs_map[parent]) > 0:
                        added_child = False
                        for child in childs_map[parent]:
                            if df_comp.loc[child,feature] > min_value:
                                parents_aux.append(child)
                                added_child = True
                        if not added_child:
                            selected_comps.append(parent)
                    else:
                        selected_comps.append(parent)
                parents = parents_aux

        elif selection_method == 'max_persistence' or selection_method == 'max_size':
            if selection_method == 'max_persistence':
                feature = 'died_at'
                max_value = max_persistence
            else:
                feature = 'size'
                max_value = max_size

            parents = [df_comp.shape[0]-1]

            while len(parents) > 0:
                parents_aux = parents.copy()
                for parent in parents:
                    parents_aux.remove(parent)
                    if df_comp.loc[parent,feature] < max_value:
                        selected_comps.append(parent)
                    else:
                        parents_aux.extend(childs_map[parent])
                parents = parents_aux
        
        else:
            raise ValueError("Selection method not recognized. Options: 'layers_down', 'n_components', 'min_persistence', 'min_size', 'max_persistence', 'max_size'")
        
        self.components_to_scale = selected_comps
        return selected_comps
    
    def _scale_component(self, component_id, i_point):
        avg_edge_comp = self.points_component_avg_edge[i_point]
        alpha = self.blowout_scalar*self.goal_length/avg_edge_comp
        alpha = min([alpha, self.max_scalar])

        comp_ids = self.components_info[component_id]['points']
        self.points_scaled[comp_ids] = True

        if self.verbose:
            print(f'Scalling component {component_id} - Scalar: {alpha:.3f}', end='')
            points = self.projections[comp_ids,:]
            area = (points[:,0].max() - points[:,0].min())*(points[:,1].max() - points[:,1].min())
            print(f' - initial area: {area:.3f}...', end='')

        t_scale = Transform(scalar=alpha)
        self.projections[comp_ids,:] = t_scale.scale(self.projections[comp_ids,:])

        if self.verbose:
            points = self.projections[comp_ids,:]
            area = (points[:,0].max() - points[:,0].min())*(points[:,1].max() - points[:,1].min())
            print(f' final area: {area:.3f}.')

        return alpha

    def _merge_components_boxes(self, c_a, c_b, i_a, i_b, d):
        # If a has a box
        if not self.points_component[i_a] is None:

            # If b does not have a box -> Join b to a's box
            if self.points_component[i_b] is None:
                a_box_id = self.points_component[i_a]
                self.points_component[list(c_b)] = a_box_id
                self.components_info[a_box_id]['points'].extend(list(c_b))
                self.components_info[a_box_id]['size'] += len(c_b)
                self.components_info[a_box_id]['persistence'] = d
                self.components_info[a_box_id]['children'] += 1

            # If b has a box -> Create parent box
            else:
                parent_box_id = len(self.components_info)
                self.components_info.append({})
                self.components_info[parent_box_id]['id'] = parent_box_id
                self.components_info[parent_box_id]['points'] = c_a.copy()
                self.components_info[parent_box_id]['points'].extend(c_b)
                self.components_info[parent_box_id]['size'] = len(c_a)+len(c_b)
                self.components_info[parent_box_id]['persistence'] = d
                self.components_info[parent_box_id]['created_at'] = d
                self.components_info[parent_box_id]['children'] = 2

                a_box_id = self.points_component[i_a]
                self.components_info[a_box_id]['parent'] = parent_box_id
                self.components_info[a_box_id]['died_at'] = d
                self.components_info[a_box_id]['persistence'] = d
                self.components_info[a_box_id]['average_edge'] = self.points_component_avg_edge[i_a]

                b_box_id = self.points_component[i_b]
                self.components_info[b_box_id]['parent'] = parent_box_id
                self.components_info[b_box_id]['died_at'] = d
                self.components_info[b_box_id]['persistence'] = d
                self.components_info[b_box_id]['average_edge'] = self.points_component_avg_edge[i_b]
                
                self.points_component[list(c_a)] = parent_box_id
                self.points_component[list(c_b)] = parent_box_id

                # Scale a and b
                if a_box_id in self.components_to_scale:
                    alpha_a = self._scale_component(a_box_id, i_a)
                    self.components_info[a_box_id]['alpha'] = alpha_a
                if b_box_id in self.components_to_scale:
                    alpha_b = self._scale_component(b_box_id, i_b)
                    self.components_info[b_box_id]['alpha'] = alpha_b

        # If a does not have a box
        else:
            # If b has a box -> Join a to b's box
            if not self.points_component[i_b] is None:
                b_box_id = self.points_component[i_b]
                self.points_component[list(c_a)] = b_box_id
                self.components_info[b_box_id]['points'].extend(list(c_a))
                self.components_info[b_box_id]['size'] += len(c_a)
                self.components_info[b_box_id]['persistence'] = d
                self.components_info[b_box_id]['children'] += 1

            # If none has box
            else:
                # If a and b merged is bigger than min -> Create box
                if len(c_a)+len(c_b) >= self.min_points_component:
                    new_box_id = len(self.components_info)
                    self.components_info.append({})
                    self.components_info[new_box_id]['id'] = new_box_id
                    self.components_info[new_box_id]['points'] = c_a.copy()
                    self.components_info[new_box_id]['points'].extend(c_b)
                    self.components_info[new_box_id]['size'] = len(c_a)+len(c_b)
                    self.components_info[new_box_id]['persistence'] = d
                    self.components_info[new_box_id]['created_at'] = d
                    self.components_info[new_box_id]['children'] = 0

                    self.points_component[list(c_a)] = new_box_id
                    self.points_component[list(c_b)] = new_box_id
    
    def _get_components(self):
        for i in range(len(self.sorted_edges)):
            # Get points from the edge
            i_a, i_b = self.sorted_edges[i][0], self.sorted_edges[i][1]

            # Distance between points
            d = self.sorted_edges[i][2]

            if self.max_edge_length!=-1 and d > self.max_edge_length:
                if self.verbose:
                    print(f'[INFO] Max edge length hit. Distance: {d} | max_edge_length: {self.max_edge_length}')
                self.next_dist = d
                break

            # Get components the points belong to
            root_a, root_b = self.uf.find(i_a), self.uf.find(i_b)
            c_a = self.uf.component[root_a].points_ids.copy()
            c_b = self.uf.component[root_b].points_ids.copy()

            # Merge boxes and scale if necessary
            self._merge_components_boxes(c_a, c_b, i_a, i_b, d)

            # Get points from edge and components
            p_a, p_b = self.projections[i_a,:], self.projections[i_b,:]
            proj_c_a, proj_c_b = self.projections[list(c_a)], self.projections[list(c_b)]

            simplices_a = self.uf.component[root_a].hull_simplices
            simplices_b = self.uf.component[root_b].hull_simplices

            # Rotate the first to be the topmost
            proj_c_a, edge_t = self._rotate_component(c_a, p_a, simplices_a, direction='top')
            # Rotate the second to be the bottomost
            proj_c_b, edge_b = self._rotate_component(c_b, p_b, simplices_b, direction='bottom')

            # Translate components
            proj_c_a = self._translate_component(proj_c_a, edge_t, to_point=[0,0])
            proj_c_b = self._translate_component(proj_c_b, edge_b, to_point=[0,d])

            self.projections[list(c_a), :] = proj_c_a
            self.projections[list(c_b), :] = proj_c_b

            avg_edge_a = self.points_component_avg_edge[i_a]
            avg_edge_b = self.points_component_avg_edge[i_b]
            avg_edge_merged = (len(c_a)*avg_edge_a+len(c_b)*avg_edge_b+d)/(len(c_a)+len(c_b)+1)
            self.points_component_avg_edge[list(c_a)] = avg_edge_merged
            self.points_component_avg_edge[list(c_b)] = avg_edge_merged

            # Merge components 
            self.uf.union(i_a, i_b, self.projections)

        if i == len(self.sorted_edges)-1:
            if self.verbose:
                print(f'[INFO] Number of edges hit. Edges processed: {i}')
            self.next_dist = d

        return self.projections

    def get_scaled_hulls(self):
        for c in self.components_to_scale:
            c_points_id = self.components_info[c]['points']
            c_hull = get_hull(self.projections[list(c_points_id),:])
            self.components_info[c]['hull'] = c_hull

        return self.components_info

    def fit_transform(self, X:np.ndarray):
        """Fit X into an embedded space and return that transformed output.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)

        Returns
        -------
        X_new : ndarray of shape (n_samples, n_components)
            Embedding of the training data in two-dimensional space.
        """
         
        n = len(X)

        self.projections = np.zeros(shape=(n, 2), dtype=np.float32)
        self.uf = UnionFind(n)
        self.points_component = np.array([None for i in range(n)])
        self.points_component_avg_edge = np.zeros(n)
        self.components_info = []
        self.points_scaled = np.zeros(shape=(n), dtype=bool)

        if self.mst is None:
            self.mst = self._compute_mst(X)

        if self.sorted_edges is None:
            self.sorted_edges = self._compute_ordered_edges(self.mst)

        if self.goal_length==-1:
            biggest_edge = self.sorted_edges[-1][2]
            self.goal_length = biggest_edge

        if len(self.components_to_scale) == 0:
            self.components_to_scale = self._get_component_to_scale(X=X,
                                                                    selection_method=self.selection_method,
                                                                    k_components=self.k_components,
                                                                    min_persistence=self.min_persistence_selection,
                                                                    min_size=self.min_size_selection,
                                                                    max_persistence=self.max_persistence_selection,
                                                                    max_size=self.max_size_selection
                                                                   )

        self._get_components()

        self.get_scaled_hulls()

        return self.projections
    